"""
Extract "upstream" and "product" from Python notebooks
"""
import ast
import parso
from collections.abc import Mapping

from ploomber.static_analysis.abc import NotebookExtractor, Extractor


class PythonNotebookExtractor(NotebookExtractor):
    def extract_upstream(self):
        return extract_upstream_assign(self.parameters_cell)

    def extract_product(self):
        """
        Extract "product" from a Python code string
        """
        product_found, product = extract_variable(self.parameters_cell,
                                                  'product')

        if not product_found or product is None:
            raise ValueError("Couldn't extract 'product' "
                             f"from code: {self.parameters_cell!r}")
        else:
            return product


def get_key_value(node):
    return node.value.id if hasattr(node.value, 'id') else None


def get_constant(node):
    # python 3.9 node.slice.value for earlier versions
    if isinstance(node.slice, ast.Constant):
        return node.slice
    # for earlier versions (if the attribute doesn't exist we are dealing with
    # the upstream[some_name]) case
    elif hasattr(node.slice, 'value'):
        return node.slice.value


def get_value(node):
    constant = get_constant(node)
    # .s for python 3.7 and 3.7, .value for 3.8 and newer
    return constant.s if hasattr(constant, 's') else constant.value


class PythonCallableExtractor(Extractor):
    def extract_upstream(self):
        """
        Extract keys requested to an upstream variable (e.g. upstream['key'])
        """
        module = ast.parse(self.code)
        return {
            get_value(node)
            for node in ast.walk(module)
            if isinstance(node, ast.Subscript) and get_key_value(node) ==
            'upstream' and isinstance(get_constant(node), ast.Str)
        } or None

    def extract_product(self):
        raise NotImplementedError('Extract product is not implemented '
                                  'for python callables')


def extract_variable(code_str, name):
    """
    Get the value assigned to a variable with name "name" by passing a code
    string
    """
    variable_found = False
    value = None

    for stmt in _iterate_assignments(code_str):
        if hasattr(stmt, 'get_defined_names'):
            defined = stmt.get_defined_names()

            if len(defined) == 1 and defined[0].value == name:
                variable_found = True
                value = eval(stmt.children[2].get_code())

    return variable_found, value


def _iterate_assignments(code_str):
    p = parso.parse(code_str)

    for ch in p.children:
        # FIXME: this works but we should find out what's the difference
        # between these two and if they are the only two valid cases
        if ch.type in ['simple_stmt', 'expr_stmt']:
            if ch.type == 'simple_stmt':
                stmt = ch.children[0]
            elif ch.type == 'expr_stmt':
                stmt = ch

            yield stmt


def extract_upstream_assign(cell_code):
    """
    Infer dependencies from a single Python cell. Looks for a cell that
    defines an upstream variable which must be either a dictionary or None
    """
    upstream_found, upstream = extract_variable(cell_code, 'upstream')

    if not upstream_found:
        raise ValueError("Could not parse a valid 'upstream' variable "
                         "in the 'parameters' cell with code:\n'%s'\n"
                         "If the notebook does not have dependencies add "
                         "upstream = None" % cell_code)
    else:
        valid_types = (Mapping, list, tuple, set)
        if not (isinstance(upstream, valid_types) or upstream is None):
            raise ValueError("Found an upstream variable but it is not a "
                             "valid type (dictionary, list, tuple set or None "
                             ", got '%s' type from code:\n"
                             "'%s'" % (type(upstream), cell_code))
        elif isinstance(upstream, valid_types):
            return set(upstream)
        else:
            return None



"""
Function for testing testing SQL relations
"""
from typing import Union, List

from jinja2 import Template


def nulls_in_columns(client, cols: Union[str, List[str]], product):
    """Check if any column has NULL values, returns bool

    Parameters
    ----------
    client
        Database client
    cols
        Column(s) to check
    product
        The relation to check

    Returns
    -------
    bool
        True if there is at least one NULL in any of the columns
    """
    # NOTE: SELECT EXISTS does not work on oracle
    # it can be SELECT 1 FROM EXISTS(...) dual (dual is a system table
    # it always exists). Should we support it?
    sql = Template("""
    SELECT EXISTS(
        SELECT * FROM {{product}}
        WHERE {{cols | join(' is null or ') }} is null
    )
    """).render(cols=cols, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = bool(cur.fetchone()[0])
    cur.close()

    return output


def distinct_values_in_column(client, col: str, product):
    """Get distinct values in a column

    Parameters
    ----------
    client
        Database client
    col
        Column to check
    product
        The relation to check

    Returns
    -------
    set
       Distinct values in column
    """

    sql = Template("""
    SELECT DISTINCT {{col}} FROM {{product}}
    """).render(col=col, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = cur.fetchall()
    cur.close()

    return set(o[0] for o in output)


def range_in_column(client, col: str, product):
    """Get range for a column

    Parameters
    ----------
    client
        Database client
    cols
        Column to check
    product
        The relation to check

    Returns
    -------
    tuple
        (minimum, maximum) values
    """
    sql = Template("""
    SELECT MIN({{col}}), MAX({{col}}) FROM {{product}}
    """).render(col=col, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = cur.fetchone()
    cur.close()

    return output


def exists_row_where(client, criteria: str, product):
    """
    Check whether at least one row exists matching the criteria

    Parameters
    ----------
    client
        Database client
    criteria
        Criteria to evaluate (passed as argument to a WHERE clause)
    product
        The relation to check

    Notes
    -----
    Runs a ``SELECT EXISTS (SELECT * FROM {{product}} WHERE {{criteria}})``
    query

    Returns
    -------
    bool
        True if exists at least one row matching the criteria
    """
    sql = Template("""
    SELECT EXISTS(
        SELECT *
        FROM {{product}}
        WHERE {{criteria}}
    )
    """).render(product=product, criteria=criteria)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = bool(cur.fetchone()[0])
    cur.close()

    return output



import copy as copy_module
from collections import abc


class Params(abc.MutableMapping):
    """
    Read-only mapping to represent params passed in Task constructor. It
    initializes with a copy of the passed dictionary. It verifies that the
    dictionary does not have a key "upstream" nor "product" because they'd
    clash with the ones added upon Task rendering
    """
    def __init__(self, params=None):
        if params is None:
            self._dict = {}
        else:
            if not isinstance(params, abc.Mapping):
                raise TypeError('Params must be initialized '
                                f'with a mapping, got: {params!r} '
                                f'({type(params).__name__!r})')

            if 'upstream' in params:
                raise ValueError('Task params cannot be initialized with an '
                                 '"upstream" key as it automatically added '
                                 'upon rendering')

            if 'product' in params:
                raise ValueError('Task params cannot be initialized with a '
                                 '"product" key as it automatically added '
                                 'upon rendering')

            self._dict = copy_module.copy(params)

    @classmethod
    def _from_dict(cls, params, copy=True):
        """
        Private API for initializing Params objects with arbitrary dictionary
        """
        obj = cls(params=None)

        if copy:
            obj._dict = copy_module.copy(params)
        else:
            obj._dict = params

        return obj

    def _setitem(self, key, value):
        """Private method for updating the underlying data
        """
        self._dict[key] = value

    def to_dict(self):
        # NOTE: do we need this?
        return copy_module.copy(self._dict)

    def to_json_serializable(self, params_only=False):
        """
        Converts params into a dictionary

        Parameters
        ----------
        params_only : bool, default=False
            If True, it only returns user params, excluding 'upstream' and
            'product'
        """
        out = self.to_dict()

        if params_only:
            out.pop('product', None)
            out.pop('upstream', None)
        elif 'upstream' in out:
            out['upstream'] = out['upstream'].to_json_serializable()

        return out

    def __getitem__(self, key):
        try:
            return self._dict[key]
        except KeyError:
            raise KeyError('Cannot obtain Task param named '
                           '"{}", declared params are: {}'.format(
                               key, list(self._dict.keys())))

    def __setitem__(self, key, value):
        raise RuntimeError('Task params are read-only, if you need a copy'
                           ' use Params.to_dict() (returns a shallow copy)'
                           ' of the underlying dictionary')

    def __iter__(self):
        for name in self._dict.keys():
            yield name

    def __len__(self):
        return len(self._dict)

    def __str__(self):
        return str(self._dict)

    def __repr__(self):
        return 'Params({})'.format(repr(self._dict))

    def get(self, key):
        return self._dict.get(key)

    def __delitem__(self, key):
        del self._dict[key]



import abc


class AbstractPlaceholder(abc.ABC):
    """
    Placeholder are objects that can eventually be converted to strings using
    str(placeholder), if this operator is used before they are fully resolved
    (e.g. they need to render first), they will raise an error.

    repr(placeholder) is always safe to use and it will return the rendered
    version if available, otherwise, it will show the raw string with
    tags, the representation is shortened if needed.

    Placeholders are mostly used inside Source objects, but are sometims
    also used to give Products placeholding features (e.g.
    by directly using Placeholder or SQLRelationPlaceholder)
    """
    @abc.abstractmethod
    def __str__(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def __repr__(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def render(self, params):
        pass  # pragma: no cover



import warnings

from ploomber.products import Product
from ploomber.placeholders.placeholder import Placeholder
from ploomber.exceptions import SourceInitializationError
from ploomber import static_analysis
from ploomber.static_analysis.string_ import StringExtractor
from ploomber.sources import docstring, abc


class PlaceholderSource(abc.Source):
    """
    Source concrete class for the ones that use a Placeholder
    """
    def __init__(self, value, hot_reload=False):
        self._primitive = value
        # rename, and make it private
        self._placeholder = Placeholder(value, hot_reload=hot_reload)
        self._post_init_validation(self._placeholder)

    @property
    def primitive(self):
        return self._primitive

    # TODO: rename to params
    @property
    def variables(self):
        return self._placeholder.variables

    def render(self, params):
        if params.get('upstream'):
            with params.get('upstream'):
                self._placeholder.render(params)
        else:
            self._placeholder.render(params)

        self._post_render_validation(str(self._placeholder), params)
        return self

    @property
    def loc(self):
        return self._placeholder.path

    def __str__(self):
        return str(self._placeholder)

    def __repr__(self):
        repr_ = "{}({})".format(
            type(self).__name__, self._placeholder.best_repr(shorten=True))

        if self._placeholder.path:
            repr_ += ' Loaded from: {}'.format(self._placeholder.path)

        return repr_

    @property
    def name(self):
        if self._placeholder.path is not None:
            # filename without extension(e.g., plot.py -> plot)
            return self._placeholder.path.stem


class SQLSourceMixin:
    @property
    def doc(self):
        return docstring.extract_from_sql(
            self._placeholder.best_repr(shorten=False))

    @property
    def extension(self):
        return 'sql'

    def extract_product(self):
        return static_analysis.sql.SQLExtractor(
            self._placeholder).extract_product()

    def extract_upstream(self):
        return static_analysis.sql.SQLExtractor(
            self._placeholder).extract_upstream()


class SQLScriptSource(SQLSourceMixin, PlaceholderSource):
    """
    A SQL (templated) script, it is expected to make a persistent change in
    the database (by using the CREATE statement), its validation verifies
    that, if no persistent changes should be validated use SQLQuerySource
    instead

    An object that represents SQL source, if a pathlib.Path object is passed,
    its contents are read and interpreted as the placeholder's content

    Notes
    -----
    This is really just a Placeholder object that stores its rendered
    version in the same object and raises an Exception if attempted. It also
    passes some of its attributes
    """
    def __init__(self, value, hot_reload=False, split_source=None):
        self._split_source = split_source
        super().__init__(value, hot_reload)

    def _post_init_validation(self, value):
        if not value.needs_render:
            example = 'CREATE TABLE {{product}} AS (SELECT * FROM ...)'
            raise SourceInitializationError(
                f'Error initializing {self!r}: '
                'The {{product}} placeholder is required. '
                f'Example: {example!r}')

        # FIXME: validate {{product}} exists, does this also catch
        # {{product['key']}} ?

    def _post_render_validation(self, rendered_value, params):
        """Analyze code and warn if issues are found
        """
        if 'product' in params:
            inferred_relations = set(
                static_analysis.sql.created_relations(
                    rendered_value, split_source=self._split_source))

            if isinstance(params['product'], Product):
                product_relations = {params['product']}
            else:
                # metaproduct
                product_relations = {p for p in params['product']}

            inferred_n = len(inferred_relations)
            actual_n = len(product_relations)

            # warn is sql code will not create any tables/views
            if not inferred_n:
                warnings.warn('It appears that your script will not create '
                              'any tables/views but the product parameter is '
                              f'{params["product"]!r}')

            # warn if the number of CREATE statements does not match the
            # number of products
            elif inferred_n != actual_n:
                warnings.warn('It appears that your script will create '
                              f'{inferred_n} relation(s) but you declared '
                              f'{actual_n} product(s): {params["product"]!r}')
            elif inferred_relations != product_relations:
                warnings.warn('It appears that your script will create '
                              f'relations {inferred_relations!r}, '
                              'which doesn\'t match products: '
                              f'{product_relations!r}. Make sure schema, '
                              'name and kind (table or view) match')

    def render(self, params):
        # FIXME: inefficient, initialize once and only update if needed
        # (i.e. hot reload is on)
        extracted = static_analysis.sql.SQLExtractor(
            self._placeholder).extract_product(raise_if_none=False)
        # the code itself might already define the product, no need to pass it
        # TODO: verify that the product passed and the one defined are the same
        if extracted is not None:
            del params['product']
            params._setitem(type(extracted).__name__, type(extracted))

        return super().render(params)


class SQLQuerySource(SQLSourceMixin, PlaceholderSource):
    """
    Templated SQL query, it is not expected to make any persistent changes in
    the database (in contrast with SQLScriptSource), so its validation is
    different
    """

    # TODO: validate this is a SELECT statement
    # a query needs to return a result, also validate that {{product}}
    # does not exist in the template, instead of just making it optional

    def render(self, params):
        if params.get('upstream'):
            with params.get('upstream'):
                self._placeholder.render(params, optional=['product'])
        else:
            self._placeholder.render(params, optional=['product'])

        self._post_render_validation(str(self._placeholder), params)

    def _post_render_validation(self, rendered_value, params):
        pass

    def _post_init_validation(self, value):
        pass


class GenericSource(PlaceholderSource):
    """
    Generic source, the simplest type of source, it does not perform any kind
    of parsing nor validation

    Prameters
    ---------
    value: str
        The value for this source


    Notes
    -----
    value is directly passed to ploomber.templates.Placeholder, which means
    pathlib.Path objects are read and str are converted to jinja2.Template
    objects, for more details see Placeholder documentation
    """
    def __init__(self, value, hot_reload=False, optional=None, required=None):
        self._primitive = value
        # rename, and make it private
        self._placeholder = Placeholder(value,
                                        hot_reload=hot_reload,
                                        required=required)
        self._post_init_validation(self._placeholder)
        self._optional = optional
        self._required = required

    def render(self, params):
        if params.get('upstream'):
            with params.get('upstream'):
                self._placeholder.render(params, optional=self._optional)
        else:
            self._placeholder.render(params, optional=self._optional)

        self._post_render_validation(str(self._placeholder), params)
        return self

    @property
    def doc(self):
        return None

    @property
    def extension(self):
        return None

    def _post_render_validation(self, rendered_value, params):
        pass

    def _post_init_validation(self, value):
        pass

    def extract_upstream(self):
        return StringExtractor(self._placeholder._raw).extract_upstream()


class FileSource(GenericSource):
    """
    A source that represents a path to a file, similar to GenericSource,
    but it casts the value argument to str, hence pathlib.Path will be
    interpreted as literals.

    This source is utilized by Tasks that move/upload files.

    """
    def __init__(self, value, hot_reload=False):
        # hot_reload does not apply here, ignored
        value = str(value)
        super().__init__(Placeholder(value))

    def render(self, params):
        if params.get('upstream'):
            with params.get('upstream'):
                self._placeholder.render(params,
                                         optional=['product', 'upstream'])
        else:
            self._placeholder.render(params, optional=['product', 'upstream'])

        self._post_render_validation(str(self._placeholder), params)

    def _post_render_validation(self, rendered_value, params):
        pass

    def _post_init_validation(self, value):
        pass

    def extract_upstream(self):
        return StringExtractor(self._placeholder._raw).extract_upstream()


class EmptySource(abc.Source):
    """A source that does not do anything, used for sourceless tasks
    """
    def __init__(self, primitive, **kwargs):
        pass

    @property
    def primitive(self):
        pass

    @property
    def variables(self):
        pass  # pragma: no cover

    def render(self, params):
        pass

    @property
    def loc(self):
        pass  # pragma: no cover

    def __str__(self):
        return 'EmptySource'

    @property
    def doc(self):
        pass  # pragma: no cover

    @property
    def extension(self):
        pass  # pragma: no cover

    def _post_render_validation(self, rendered_value, params):
        pass  # pragma: no cover

    def _post_init_validation(self, value):
        pass  # pragma: no cover

    @property
    def name(self):
        pass  # pragma: no cover



import os
from functools import wraps
import sys

try:
    # optional dependency
    import ipdb as pdb
except ModuleNotFoundError:
    import pdb

import click

from ploomber.io import TerminalWriter
from ploomber.exceptions import DAGBuildError, DAGRenderError, BaseException
from ploomber.executors import _format

# TODO: there are two types of cli commands: the ones that execute user's
# code (ploomber build/task) and the ones that parse a dag/task but do not
# execute it. For the former, we want to capture errors and display them with
# colors so it's easier for the user to understand what went wrong with their
# code. For the latter, the errors are raise by us, hence, we only need to
# print the message and exit. Currently, all CLI end points (except ploomber
# nb) are decorated with @cli_endpoint but we should change it to
# @command_endpoint


def cli_endpoint(fn):
    """
    Decorator for command line endpoints that execute dags or tasks. It runs
    the decorated function, captures exception (if any), sends a colored
    traceback to standard error and exits with code 1.

    Notes
    -----
    This will hide the traceback when raising subclasses of
    ploomber.exeptions.BaseException. To display the traceback, set the
    PLOOMBER_DEBUG variable to true. To start a post-mortem session, set
    PLOOMBER_POST_MORTEM to true.

    Functions decorated with this must be called with keyword arguments

    Call some_endpoint(catch_exception=False) to disable this behavior (e.g.
    for testing)
    """
    @wraps(fn)
    def wrapper(catch_exception=True, **kwargs):
        if os.environ.get('PLOOMBER_DEBUG'):
            catch_exception = False

        if catch_exception:
            try:
                fn(**kwargs)
            # these already color output
            except (DAGBuildError, DAGRenderError) as e:
                error = str(e)
                color = False
            except BaseException as e:
                click.secho(e.get_message(), file=sys.stderr, fg='red')
                sys.exit(1)
            except Exception as e:
                error = _format.exception(e)
                color = True
            else:
                error = None

            if error:
                if color:
                    tw = TerminalWriter(file=sys.stderr)
                    tw._write_source(error.splitlines())
                else:
                    print(error, file=sys.stderr)

                sys.exit(1)
        else:
            if os.environ.get('PLOOMBER_POST_MORTEM'):
                try:
                    fn(**kwargs)
                except Exception:
                    _, _, tb = sys.exc_info()
                    pdb.post_mortem(tb)
            else:
                fn(**kwargs)

    return wrapper


# FIXME: capture only certain types of exceptions. If it's something we didn't
# raise, we'd like to see the full traceback
def command_endpoint(fn):
    """
    Decorator for command line endpoints that only parse dags or tasks but do
    not execute them. If it tails, it prints error message to stderror, then
    calls with exit code 1.
    """
    @wraps(fn)
    def wrapper(**kwargs):
        try:
            fn(**kwargs)
        except BaseException as e:
            click.secho(e.get_message(), file=sys.stderr, fg='red')
            sys.exit(1)
        except Exception as e:
            print(f'Error: {e}', file=sys.stderr)
            sys.exit(1)

    return wrapper



import os
import contextlib
import tempfile
from pathlib import Path
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed

from ploomber.exceptions import DAGWithDuplicatedProducts
from ploomber.products.metaproduct import MetaProduct
from ploomber.products import File
from ploomber.io import pretty_print


def _generate_error_message_pair(key, value):
    return f'* {key!r} generated by tasks: {pretty_print.iterable(value)}'


def _generate_error_message(duplicated):
    return '\n'.join(
        _generate_error_message_pair(key, value)
        for key, value in duplicated.items())


def check_duplicated_products(dag):
    """
    Raises an error if more than one task produces the same product.

    Note that this relies on the __hash__ and __eq__ implementations of
    each Product to determine whether they're the same or not. This
    implies that a relative File and absolute File pointing to the same file
    are considered duplicates and SQLRelations (in any of its flavors) are
    the same when they resolve to the same (schema, name, type) tuple
    (i.e., client is ignored), this because when using the generic SQLite
    backend for storing SQL product metadata, the table only relies on schema
    and name to retrieve metadata.
    """
    prod2tasknames = defaultdict(lambda: [])

    for name in dag._iter():
        product = dag[name].product

        if isinstance(product, MetaProduct):
            for p in product.products:
                prod2tasknames[p].append(name)
        else:
            prod2tasknames[product].append(name)

    duplicated = {k: v for k, v in prod2tasknames.items() if len(v) > 1}

    if duplicated:
        raise DAGWithDuplicatedProducts(
            'Tasks must generate unique products. '
            'The following products appear in more than '
            f'one task:\n{_generate_error_message(duplicated)}')


def flatten_products(elements, require_file_client=True):
    flat = []

    for prod in elements:
        if isinstance(prod, MetaProduct):
            flat.extend([p for p in prod if isinstance(p, File) and p.client])
        elif (isinstance(prod, File) and prod.client
              and require_file_client) or (isinstance(prod, File)
                                           and not require_file_client):
            flat.append(prod)

    return flat


def fetch_remote_metadata_in_parallel(dag):
    """Fetches remote metadta in parallel from a list of Files
    """
    files = flatten_products(dag[t].product for t in dag._iter()
                             if isinstance(dag[t].product, File)
                             or isinstance(dag[t].product, MetaProduct))

    if files:
        with ThreadPoolExecutor(max_workers=64) as executor:
            future2file = {
                executor.submit(file._remote._fetch_remote_metadata): file
                for file in files
            }

            for future in as_completed(future2file):
                exception = future.exception()

                if exception:
                    local = future2file[future]
                    raise RuntimeError(
                        'An error occurred when fetching '
                        f'remote metadata for file {local!r}') from exception


@contextlib.contextmanager
def _path_for_plot(path_to_plot, fmt):
    """Context manager to manage DAG.plot

    Parameters
    ----------
    path_to_plot : str
        Where to store the plot. If 'embed', It returns a temporary empty file
        otherwise and deletes it when exiting. Otherwise, it just passes the
        value
    """
    if path_to_plot == 'embed':
        fd, path = tempfile.mkstemp(suffix=f'.{fmt}')
        os.close(fd)
    else:
        path = str(path_to_plot)

    try:
        yield path
    finally:
        if path_to_plot == 'embed':
            Path(path).unlink()


def iter_file_products(dag):
    for t in dag._iter():
        product = dag[t].product

        if isinstance(product, File):
            yield product
        elif isinstance(product, MetaProduct):
            for prod in product:
                if isinstance(prod, File):
                    yield prod


def _get_parent_from_product(product):
    path = Path(str(product._identifier))
    current = Path().resolve()

    if path.is_absolute():
        try:
            # products loaded from pipeline.yaml are always absolute, so try
            # to see if they're relative to the current directory
            path = path.relative_to(current)
        except ValueError as e:
            raise ValueError('Absolute product paths '
                             f'are not supported: {str(path)!r}') from e

    return str(path.parent)


def extract_product_prefixes(dag):
    files = set(_get_parent_from_product(p) for p in iter_file_products(dag))

    return sorted(files)



import json
import pickle
from pathlib import Path
from functools import wraps
from inspect import signature

try:
    from joblib import load as joblib_load
except ModuleNotFoundError:
    joblib_load = None

try:
    from cloudpickle import load as cloudpickle_load
except ModuleNotFoundError:
    cloudpickle_load = None
try:
    import pandas as pd
except ModuleNotFoundError:
    pd = None

from ploomber.products import MetaProduct
from ploomber.io.serialize import _build_extension_mapping_final

_EXTERNAL = {
    False: None,
    True: pickle.load,
    'pickle': pickle.load,
    'joblib': joblib_load,
    'cloudpickle': cloudpickle_load,
}


def _txt2str(product):
    return Path(product).read_text()


def _json2obj(product):
    return json.loads(Path(product).read_text())


def _csv2df(product):
    if pd is None:
        raise ModuleNotFoundError('Error using .csv default unserializer: '
                                  'pandas is not installed')

    return pd.read_csv(product)


def _parquet2df(product):
    if pd is None:
        raise ModuleNotFoundError('Error using .parquet default unserializer: '
                                  'pandas is not installed')

    return pd.read_parquet(product)


_DEFAULTS = {
    '.txt': _txt2str,
    '.json': _json2obj,
    '.csv': _csv2df,
    '.parquet': _parquet2df,
}


def unserializer(extension_mapping=None,
                 *,
                 fallback=False,
                 defaults=None,
                 unpack=False):
    """Decorator for unserializing functions

    Parameters
    ----------
    extension_mapping : dict, default=None
        An extension -> function mapping. Calling the decorated function with a
        File of a given extension will use the one in the mapping if it exists,
        e.g., {'.csv': from_csv, '.json': from_json}.

    fallback : bool or str, default=False
        Determines what method to use if extension_mapping does not match the
        product to unserialize. Valid values are True (uses the pickle module),
        'joblib', and 'cloudpickle'. If you use any of the last two, the
        corresponding moduel must be installed. If this is enabled, the
        body of the decorated function is never executed. To turn it off
        pass False.

    defaults : list, default=None
        Built-in unserializing functions to use. Must be a list with any
        combinations of values: '.txt', '.json', '.csv', '.parquet'.
        Unserializing .txt, returns a string, for .json returns any
        JSON-unserializable object (e.g., a list or a dict), .csv and
        .parquet return a pandas.DataFrame. If using .parquet, a parquet
        library must be installed (e.g., pyarrow). If extension_mapping
        and defaults contain overlapping keys, an error is raises

    unpack : bool, default=False
        If True and the task product points to a directory, it will call
        the unserializer one time per file in the directory. The unserialized
        object will be a dictionary where keys are the filenames and values
        are the unserialized objects. Note that this isn't recursive, it only
        looks at files that are immediate children of the product directory.
    """
    def _unserializer(fn):
        extension_mapping_final = _build_extension_mapping_final(
            extension_mapping, defaults, fn, _DEFAULTS, 'unserializer')

        try:
            unserializer_fallback = _EXTERNAL[fallback]
        except KeyError:
            error = True
        else:
            error = False

        if error:
            raise ValueError(f'Invalid fallback argument {fallback!r} '
                             f'in function {fn.__name__!r}. Must be one of '
                             "True, 'joblib', or 'cloudpickle'")

        if unserializer_fallback is None and fallback in {
                'cloudpickle', 'joblib'
        }:
            raise ModuleNotFoundError(
                f'Error unserializing with function {fn.__name__!r}. '
                f'{fallback} is not installed')

        n_params = len(signature(fn).parameters)
        if n_params != 1:
            raise TypeError(f'Expected unserializer {fn.__name__!r} '
                            f'to take 1 argument, but it takes {n_params!r}')

        @wraps(fn)
        def wrapper(product):
            if isinstance(product, MetaProduct):
                return {
                    key: _unserialize_product(value, extension_mapping_final,
                                              fallback, unserializer_fallback,
                                              fn, unpack)
                    for key, value in product.products.products.items()
                }

            else:
                return _unserialize_product(product, extension_mapping_final,
                                            fallback, unserializer_fallback,
                                            fn, unpack)

        return wrapper

    return _unserializer


@unserializer(fallback=True)
def unserializer_pickle(product):
    """An unserializer that unpickles everything
    """
    # this should never execute
    raise RuntimeError('Error when unserializing with pickle module')


def _make_unserializer(fn):
    def _unserialize(product):
        with open(product, 'rb') as f:
            obj = fn(f)

        return obj

    return _unserialize


def _unserialize_product(product, extension_mapping, fallback,
                         unserializer_fallback, fn, unpack):

    if unpack and Path(product).is_dir():
        out = {}

        for path in Path(product).glob('*'):
            unserializer = _determine_unserializer(path, extension_mapping,
                                                   fallback,
                                                   unserializer_fallback, fn)

            out[path.name] = unserializer(path)

        return out

    # treat product as a single file...
    else:
        unserializer = _determine_unserializer(product, extension_mapping,
                                               fallback, unserializer_fallback,
                                               fn)
        return unserializer(product)


def _determine_unserializer(product, extension_mapping, fallback,
                            unserializer_fallback, fn):
    suffix = Path(product).suffix

    if extension_mapping and suffix in extension_mapping:
        return extension_mapping[suffix]
    elif fallback:
        unserializer = _make_unserializer(unserializer_fallback)
        return unserializer
    else:
        return fn



def find_cell_with_tags(nb, tags):
    """
    Find the first cell with any of the given tags, returns a dictionary
    with 'cell' (cell object) and 'index', the cell index.
    """
    tags_to_find = list(tags)
    tags_found = {}

    for index, cell in enumerate(nb['cells']):
        for tag in cell['metadata'].get('tags', []):
            if tag in tags_to_find:
                tags_found[tag] = dict(cell=cell, index=index)
                tags_to_find.remove(tag)

                if not tags_to_find:
                    break

    return tags_found


def find_cell_with_tag(nb, tag):
    """
    Find a cell with a given tag, returns a cell, index tuple. Otherwise
    (None, None)
    """
    out = find_cell_with_tags(nb, [tag])

    if out:
        located = out[tag]
        return located['cell'], located['index']
    else:
        return None, None



"""
A generic product whose metadata is saved in a given directory and
exists/delete methods are bash commands
"""
from ploomber.products.product import Product
from ploomber.products.mixins import SQLProductMixin, ProductWithClientMixin
from ploomber.products.sql import SQLiteBackedProductMixin
from ploomber.placeholders.placeholder import (Placeholder,
                                               SQLRelationPlaceholder)


# TODO: add check_product and run tests: e.g .name should return a string
# no placeholders objects or {{}}
class GenericProduct(SQLiteBackedProductMixin, ProductWithClientMixin,
                     Product):
    """
    GenericProduct is used when there is no specific Product implementation.
    Sometimes it is technically possible to write a Product implementation
    but if you don't want to do it you can use this one. Other times is is
    not possible to provide a concrete Product implementation (e.g. we
    cannot store arbitrary metadata in a Hive table). GenericProduct works
    as any other product but its metadata is stored not in the Product itself
    but in a different backend.

    Parameters
    ----------
    identifier : str
        An identifier fot this product, can contain placeholders
        (e.g. {{placeholder}})

    client : ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to *store metadata for this product*. Only required
        if no dag-level client has been declared using dag.clients[class]

    Notes
    -----
    exists does not check for product existence, just checks if metadata exists
    delete does not perform actual deletion, just deletes metadata
    """
    def __init__(self, identifier, client=None):
        super().__init__(identifier)
        self._client = client

    def _init_identifier(self, identifier):
        return Placeholder(identifier)

    def exists(self):
        # just check if there is metadata
        return self.fetch_metadata() is not None

    def delete(self, force=False):
        """Deletes the product
        """
        # just delete the metadata, we cannot do anything else
        return self._delete_metadata()

    @property
    def name(self):
        return str(self._identifier)


class GenericSQLRelation(SQLProductMixin, GenericProduct):
    """
    A GenericProduct whose identifier is a SQL relation, uses SQLite as
    metadata backend

    Parameters
    ----------
    identifier : tuple of length 3
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'

    client : ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to *store metadata for this product*. Only required
        if no dag-level client has been declared using dag.clients[class]
    """
    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind

    def __repr__(self):
        return f'{type(self).__name__}({self._identifier._raw_repr()})'

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))


class SQLRelation(SQLProductMixin, Product):
    """A product that represents a SQL relation but has no metadata

    Parameters
    ----------
    identifier : tuple of length 3
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'

    """
    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind

    def fetch_metadata(self):
        return None

    def save_metadata(self, metadata):
        pass

    def exists(self):
        return True

    def delete(self, force=False):
        pass

    def __repr__(self):
        return f'{type(self).__name__}({self._identifier._raw_repr()})'

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash((self.schema, self.name, self.kind))



import json
from pathlib import Path
from importlib.util import find_spec

import jinja2
from IPython.display import HTML

try:
    import importlib.resources as importlib_resources
except ImportError:  # pragma: no cover
    # backported
    import importlib_resources

from ploomber import resources
from ploomber.util.util import requires


def check_pygraphviz_installed():
    return find_spec("pygraphviz") is not None


def choose_backend(backend):
    """Determine which backend to use for plotting
    """
    if ((not check_pygraphviz_installed() and backend is None)
            or (backend == 'd3')):
        return 'd3'
    else:
        return 'pygraphviz'


def json_dag_parser(graph: dict):
    """Format dag dict so d3 can understand it
    """
    nodes = {}

    for task in graph["nodes"]:
        nodes[task["id"]] = task

    # change name label to products for now
    for node in nodes:
        nodes[node]["products"] = (nodes[node]["label"].replace("\n",
                                                                "").split(","))

    for link in graph["links"]:
        node_links = nodes[link["target"]].get("parentIds", [])
        node_links.append(link["source"])
        nodes[link["target"]]["parentIds"] = node_links

    return json.dumps(list(nodes.values()))


def with_d3(graph, output):
    """Generates D3 Dag html output and return output file name
    """
    template = jinja2.Template(
        importlib_resources.read_text(resources, 'dag_template.html'))

    rendered = template.render(json_data=json_dag_parser(graph=graph))
    Path(output).write_text(rendered)


@requires(['requests_html', 'nest_asyncio'],
          name='embedded HTML with D3 backend',
          pip_names=['requests-html', 'nest_asyncio'])
def embedded_html(path):
    import asyncio
    import nest_asyncio
    nest_asyncio.apply()
    return asyncio.get_event_loop().run_until_complete(
        _embedded_html(path=path))


async def _embedded_html(path):
    # https://github.com/jupyter/nbclient/blob/1d629b2bed561fde521e6408e190a8159f117ddc/nbclient/util.py
    # https://github.com/jupyter/nbclient/blob/main/requirements.txt
    from requests_html import HTML as HTML_, AsyncHTMLSession

    session = AsyncHTMLSession()
    html = HTML_(html=Path(path).read_text(), session=session)
    await html.arender()
    # ensure we close the session, otherwise this will fail on windows
    await session.close()
    return HTML(data=html.html)



from pathlib import Path
import shutil

from ploomber.util.default import find_root_recursively
from ploomber.clients.storage.abc import AbstractStorageClient
from ploomber.clients.storage.util import _resolve
from ploomber.exceptions import RemoteFileNotFound, DAGSpecInvalidError


class LocalStorageClient(AbstractStorageClient):
    """

    Parameters
    ----------
    path_to_backup_dir
        Local directory to use as backup

    path_to_project_root : str, default=None
        Path to project root. Product locations ares stored in a path relative
        to this folder. e.g. If project root is ``/my-project``, backup is
        ``/backup`` and you save a file in ``/my-project/reports/report.html``,
        it will be saved at ``/backup/reports/report.html``. If None, looks it
        up automatically and assigns it to the parent folder of the root YAML
        spec ot setup.py (if your project is a package).
    """

    def __init__(self, path_to_backup_dir, path_to_project_root=None):
        self._path_to_backup_dir = Path(path_to_backup_dir)
        self._path_to_backup_dir.mkdir(exist_ok=True, parents=True)

        if path_to_project_root:
            project_root = path_to_project_root
        else:
            try:
                project_root = find_root_recursively()
            except Exception as e:
                raise DAGSpecInvalidError(
                    f'Cannot initialize {self!r} because there '
                    'is not project root. Set one or explicitly pass '
                    'a value in the path_to_project_root argument') from e

        self._path_to_project_root = Path(project_root).resolve()

    def _remote_path(self, local):
        relative = _resolve(local).relative_to(self._path_to_project_root)
        return Path(self._path_to_backup_dir, relative)

    def _remote_exists(self, local):
        return self._remote_path(local).exists()

    def download(self, local, destination=None):
        remote = self._remote_path(local)
        destination = destination or local
        Path(destination).parent.mkdir(exist_ok=True, parents=True)

        if remote.is_file():
            # shutil.copy tries to preserve metadata, we found a race
            # condition on macOS (our CI would fail from time to time)
            # so we are now using copyfile since it doesn't copy metadata
            shutil.copyfile(remote, destination)
        elif remote.is_dir():
            shutil.copytree(remote, destination)
        else:
            raise RemoteFileNotFound('Could not download '
                                     f'{str(local)!r} using client {self}: '
                                     'No such file or directory')

    def upload(self, local):
        remote_path = self._remote_path(local)
        remote_path.parent.mkdir(exist_ok=True, parents=True)

        if Path(local).is_file():
            shutil.copy(local, remote_path)
        else:
            shutil.copytree(local, remote_path)

    def _download(self, local, remote):
        raise NotImplementedError

    def _upload(self, local):
        raise NotImplementedError

    def _is_file(self, remote):
        raise NotImplementedError

    def _is_dir(self, remote):
        raise NotImplementedError

    def __repr__(self):
        return f'{type(self).__name__}({str(self._path_to_backup_dir)!r})'



from IPython import start_ipython
from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.telemetry import telemetry


@cli_endpoint
@telemetry.log_call('interact')
def main():
    parser = CustomParser(description='Call an entry point '
                          '(pipeline.yaml or dotted path to factory)',
                          prog='ploomber interact')
    with parser:
        # this command has no static args
        pass

    dag, _ = parser.load_from_entry_point_arg()

    try:
        dag.render()
    except Exception:
        err = ('Your dag failed to render, but you can still inspect the '
               'object to debug it.\n')
        telemetry.log_api("interact-error-render",
                          metadata={
                              'type': 'dag_render_failed',
                              'exception': err,
                              'dag': dag,
                          })
        print(err)

    # NOTE: do not use embed here, we must use start_ipython, see here:
    # https://github.com/ipython/ipython/issues/8918
    start_ipython(argv=[], user_ns={'dag': dag})
